/**
  *  \file client/scr-base.cc
  *  \brief Starbase Screen
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/cscreen.h"
#include "client/marks.h"
#include "game/global.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "client/actions.h"
#include "game/ship.h"
#include "ui/listbox.h"
#include "ui/widgets.h"
#include "client/dlg-tech.h"
#include "ui/table.h"
#include "client/dlg-structure.h"
#include "client/dlg-ammo.h"
#include "client/dlg-build.h"
#include "resmgr/resid.h"

namespace {
    class WBaseSelect : public WObjectSelectionTile {
     public:
        WBaseSelect(GUniverse& univ);
        void drawData(GfxCanvas& can);
    };

    class WBaseScreen : public WControlScreen {
        SignalConnection sig_conn;
     public:
        WBaseScreen();
        ~WBaseScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void update();
    };

    class WBaseMineralTile : public WTile {
     public:
        WBaseMineralTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
    };

    class WBaseTechTile : public WTile {
     public:
        WBaseTechTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
        void addControls(UIGroup& g, ObjectHolder& h);
    };

    class WBaseOrderTile : public WTile {
     public:
        WBaseOrderTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
        void addControls(UIGroup& g, ObjectHolder& holder);
    };
}

WBaseSelect::WBaseSelect(GUniverse& univ)
    : WObjectSelectionTile(univ.getBaseType(), univ.sig_planet_changed)
{ }

void
WBaseSelect::drawData(GfxCanvas& can)
{
    GPlanet& pp = getType().getUniverse().getPlanet(getCurrent());

    int y = 0, rem = 0;
    if (pp.isMarked()) {
        drawBackground(can, GfxRect(0, 0, 15, font_heights[FONT_NORMAL]));
        getSkin().getParameters().color = standard_colors[COLOR_GREENBLACK];
        drawSelection(can, getSkin(), extent.x + 5, extent.y + 7, 1, 2);
        rem = 15;
    }
    setColor(UIColor::tc_Static);
    outTextF(can, FONT_NORMAL, rem, y, extent.w - rem, pp.getName());
    y += font_heights[FONT_NORMAL]+2;

    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             format(_("%d mc; %d kt supplies"),
                    pp.getCargo(el_Money), pp.getCargo(el_Supplies)));
    y += font_heights[FONT_NORMAL];

    /* minerals */
    outTextF(can, FONT_NORMAL, 0, y, 50, _("Minerals:"));
    getSkin().getParameters().just_x = 2;
    outTextF(can, FONT_NORMAL, 130, y, 80, format("%dN", pp.getCargo(el_Neutronium)));
    outTextF(can, FONT_NORMAL, 210, y, 80, format("%dT", pp.getCargo(el_Tritanium)));
    outTextF(can, FONT_NORMAL, 290, y, 80, format("%dD", pp.getCargo(el_Duranium)));
    outTextF(can, FONT_NORMAL, 370, y, 80, format("%dM", pp.getCargo(el_Molybdenum)));
    y += font_heights[FONT_NORMAL];

    /* tech */
    getSkin().getParameters().just_x = 0;
    outTextF(can, FONT_NORMAL, 0, y, 100, _("Tech Levels:"));
    getSkin().getParameters().just_x = 2;
    outTextF(can, FONT_NORMAL, 130, y, 30, format("%dH", pp.getBaseTech(tl_Hull)));
    outTextF(can, FONT_NORMAL, 210, y, 80, format("%dE", pp.getBaseTech(tl_Engine)));
    outTextF(can, FONT_NORMAL, 290, y, 80, format("%dB", pp.getBaseTech(tl_Beam)));
    outTextF(can, FONT_NORMAL, 370, y, 80, format("%dT", pp.getBaseTech(tl_Torpedo)));
    y += font_heights[FONT_NORMAL]+2;

    /* are we building something? */
    getSkin().getParameters().just_x = 0;
    if (int sid = pp.getShipCloningHere(getDisplayedTurn().getCurrentUniverse())) {
        outTextF(can, FONT_NORMAL, 0, y, extent.w, format(_("Cloning %s"),
                                                          getDisplayedTurn().getShip(sid).getName()));
    } else if (int hid = pp.getBaseBuildHull()) {
        outTextF(can, FONT_NORMAL, 0, y, extent.w, format(_("Building a %s"), getHull(hid).getName()));
    } else {
        outTextF(can, FONT_NORMAL, 0, y, extent.w, string_t());
    }
    y += font_heights[FONT_NORMAL];

    /* repairing something? */
    if (pp.getBaseShipyardOrder()) {
        outTextF(can, FONT_NORMAL, 0, y, extent.w,
                 format(pp.getBaseShipyardOrder() == 1 ? _("Repairing %s") : _("Recycling %s"),
                        getDisplayedTurn().getShip(pp.getBaseShipyardId()).getFullName()));
    } else {
        outTextF(can, FONT_NORMAL, 0, y, extent.w, string_t());
    }
    y += font_heights[FONT_NORMAL]+2;

    /* administrivia */
    outTextF(can, FONT_NORMAL, 0, y, extent.w, format(_("Planetary FCode: %s"), pp.getFCode()));
    y += font_heights[FONT_NORMAL];
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             format(_("SB Defense: %d; Fighters: %d; Damage: %d%%"),
                    pp.getStructures(str_BaseDefense), pp.getCargo(el_Fighters), pp.getBaseDamage()));
    y += font_heights[FONT_NORMAL];
}

int
chooseBase()
{
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    int nbases = univ.getBaseType().getCount();
    if (nbases == 0) {
        messageBox(_("You do not have any starbases. "
                     "To build starbases, use [F8] on a planet."),
                   _("Starbases"));
        return 0;
    } else if (nbases == 1) {
        return univ.getBaseType().getCurrent();
    } else {
        return WBaseSelect(univ).doStandardDialog(_("Select Starbase"));
    }
}

/**************************** WBaseMineralTile ***************************/

WBaseMineralTile::WBaseMineralTile()
    : WTile(GfxPoint(315, font_heights[FONT_NORMAL] * 4))
{ }

static void
drawBaseMineralTile(UIAbstractTableWriter& tw)
{
    GPlanet&  p = getDisplayedTurn().getBaseType().getPlanet();

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999 kt"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p.getCargo(el_Neutronium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p.getCargo(el_Tritanium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p.getCargo(el_Duranium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p.getCargo(el_Molybdenum)));

    tw.newLColumn(5);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999 mc"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d mc", p.getCargo(el_Money)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p.getCargo(el_Supplies)));

    tw.finish();
}

void
WBaseMineralTile::drawFixed(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_IgnoreVariable);
    drawBaseMineralTile(tw);
}

void
WBaseMineralTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_IgnoreFixed);
    drawBaseMineralTile(tw);
}

/***************************** WBaseTechTile *****************************/

WBaseTechTile::WBaseTechTile()
    : WTile(315, 4 * font_heights[FONT_NORMAL])
{ }

static void
drawBaseTechTile(UIAbstractTableWriter& tw)
{
    const GPlanet& p = getDisplayedTurn().getBaseType().getPlanet();

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Hulls:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Beam Weapons:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Torpedoes:"));

    tw.newRColumn(2 * font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Green, itoa(p.getBaseTech(tl_Engine)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p.getBaseTech(tl_Hull)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p.getBaseTech(tl_Beam)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p.getBaseTech(tl_Torpedo)));

    tw.newLColumn(5);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Defense:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fighters:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth(_("none")));

    tw.writeVariableCell(UIColor::tc_Green, itoa(p.getStructures(str_BaseDefense)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p.getCargo(el_Fighters)));
    if (p.getBaseDamage()) {
        tw.writeVariableCell(UIColor::tc_Red, itoa(p.getBaseDamage()) + "%");
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    tw.finish();
}

void
WBaseTechTile::drawFixed(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_IgnoreVariable);
    drawBaseTechTile(tw);
}

void
WBaseTechTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_IgnoreFixed);
    drawBaseTechTile(tw);
}

void
WBaseTechTile::addControls(UIGroup& g, ObjectHolder& h)
{
    const int r = getExtent().w;
    addControl(g, GfxRect(r-20,  0, 20, 20), h.add(new UIButton("T", 't')));
    addControl(g, GfxRect(r-20, 25, 20, 20), h.add(new UIButton("D", 'd')));
}

/***************************** WBaseOrderTile ****************************/

WBaseOrderTile::WBaseOrderTile()
    : WTile(315, 100)
{ }

void
WBaseOrderTile::drawFixed(GfxCanvas& can)
{
    setColor(UIColor::tc_Static);
    outText(can, FONT_NORMAL, 35, 55, _("Mission:"));
    outText(can, FONT_NORMAL, 35, 80, _("Build Fighters/Torpedoes"));
}

void
WBaseOrderTile::drawData(GfxCanvas& can)
{
    const GPlanet& p = getDisplayedTurn().getBaseType().getPlanet();

    /* ship */
    int hull = p.getBaseBuildHull();
    int clnid = p.getShipCloningHere(getDisplayedTurn().getCurrentUniverse());
    string_t text;

    if (hull) {
        if (clnid) {
            setColor(UIColor::tc_Red);
            text = format(_("Cloning, and building %s"),
                          getHull(hull).getName());
        } else {
            setColor(UIColor::tc_Yellow);
            text = format(_("Building a %s"),
                          getHull(hull).getName());
        }
    } else {
        if (clnid) {
            setColor(UIColor::tc_Yellow);
            text = format(_("Cloning %s"),
                          getDisplayedTurn().getShip(clnid).getName());
        } else {
            setColor(UIColor::tc_Static);
            text = _("Build a new ship");
        }
    }
    outTextF(can, FONT_NORMAL, 35, 5, extent.w-35, text);

    /* fix/recycle */
    if (p.getBaseShipyardOrder()) {
        setColor(UIColor::tc_Yellow);
        outTextF(can, FONT_NORMAL, 35, 30, extent.w-35,
                 format(p.getBaseShipyardOrder() == 1
                        ? _("Repairing %s")
                        : _("Recycling %s"),
                        getDisplayedTurn().getShip(p.getBaseShipyardId()).getFullName()));
    } else {
        setColor(UIColor::tc_Static);
        outTextF(can, FONT_NORMAL, 35, 30, extent.w-35, _("Repair/recycle a starship"));
    }

    /* mission */
    const int mx = 65 + fonts[FONT_NORMAL]->getTextWidth(_("Mission:"));
    setColor(UIColor::tc_Green);
    outTextF(can, FONT_NORMAL, mx, 55, extent.w-mx, p.getBaseMissionName());
}

void
WBaseOrderTile::addControls(UIGroup& g, ObjectHolder& h)
{
    addControl(g, GfxRect(5,  2, 20, 20), h.add(new UIButton("B", 'b')), true);
    addControl(g, GfxRect(5, 27, 20, 20), h.add(new UIButton("R", 'r')), true);
    addControl(g, GfxRect(5, 52, 20, 20), h.add(new UIButton("M", 'm')), true);
    addControl(g, GfxRect(5, 77, 20, 20), h.add(new UIButton("A", 'a')), true);
}

/****************************** WBaseScreen ******************************/

WBaseScreen::WBaseScreen()
    : WControlScreen(getDisplayedTurn().getBaseType(), 3, GPlanet::ps_DirtyHead)
{
    addTile(holder.add(new WControlScreenHeaderTile(*this, WControlScreenHeaderTile::bAuto + WControlScreenHeaderTile::bAdd,
                                                    GPlanet::ps_DirtyHead)));
    addTile(holder.add(new WBaseMineralTile()),
            _("Minerals & Funds:"));
    addTile(holder.add(new WBaseTechTile()),
            _("Technology & Defense:"));
    addTile(holder.add(new WBaseOrderTile()),
            _("Orders:"));
    addTile(holder.add(new WPlanetFCodeTile(getDisplayedTurn().getBaseType())));
    addTile(holder.add(new WPlanetBaseTile(true)));
    addChart(holder.add(new WScannerChartWidget()));

    sig_conn = getDisplayedTurn().getCurrentUniverse().
        sig_planet_changed.add(static_cast<WControlScreen*>(this),
                               &WControlScreen::updateId);
}

WBaseScreen::~WBaseScreen()
{ }

static void
doBaseXchg(WControlScreen& scr)
{
    WScanner& scanner = scr.getScanner();
    int pid = scanner.getScannedPlanetId(scanner.getScannedShipId() != 0);
    if (pid && pid != scr.getType().getCurrent()) {
        if (scr.getType().isExistant(pid))
            putCommand(cm_CS_GotoCS, 3, pid);
        else if (getDisplayedTurn().getPlanetType().isExistant(pid))
            putCommand(cm_CS_GotoCS, 2, pid);
        else
            putKey(ss_Ctrl + SDLK_F1);
    } else {
        putKey(ss_Ctrl + SDLK_F1);
    }
}

bool
WBaseScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        try {
            int pid         = getType().getCurrent();
            GUniverse& univ = getType().getUniverse();
            GPlanet& me     = univ.getPlanet(pid);
            switch (e.key.code) {
             case 'x':
                doBaseXchg(*this);
                return true;
             case 'f':
                doPlanetFCodeChange(univ, me);
                return true;
             case 'c':
             case 'u':
                doPlanetTransfer(univ, me, e.key.code == 'u');
                return true;
             case SDLK_F8:
                putCommand(cm_CS_GotoCS, 2, pid);
                return true;
             case 'm':
                doBaseChangeMission(univ, me);
                return true;
             case 'r':
                doBaseFixRecycle(univ, me, e);
                return true;
             case ss_Ctrl + 'r':
                if (int sid = me.getBaseShipyardId())
                    putCommand(cm_CS_GotoCS, 1, sid);
                return true;
             case ss_Ctrl + 'c':
                if (int sid = me.getShipCloningHere(univ))
                    putCommand(cm_CS_GotoCS, 1, sid);
                return true;
             case 't':
                doTechUpgradeDialog(univ, me);
                return true;
             case 'd':
                doPlanetBuildScreen(univ, me, 2);
                return true;
             case 'a':
                doBaseBuildAmmo(univ, me);
                return true;
             case 'b':
             {
                 WShipBuildScreenWithActor bs(me);
                 ui_root->centerWidget(bs);
                 ui_root->add(bs);
                 ui_root->run();
                 ui_root->close(bs);
                 univ.doUpdates();
                 return true;
             }

             /* base screen keys:
                's'    ship cost
                ins    sim
                alt-r  reset
                f10    choose fleet */

            }
        }
        catch (const GError& e) {
            messageBox(e.getUserError(), _("Error"));
            return true;
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

void
WBaseScreen::update()
{
    const GPlanet& planet = getDisplayedTurn().getBaseType().getPlanet();
    int max_tech = std::max(std::max(planet.getBaseTech(tl_Engine),
                                     planet.getBaseTech(tl_Hull)),
                            std::max(planet.getBaseTech(tl_Beam),
                                     planet.getBaseTech(tl_Torpedo)));
    setData(ResId(res::base, max_tech, planet.getId()),
            planet.getPos(),
            planet.getName(),
            format(_("(Id #%d)"), planet.getId()));
}

int
doBaseScreen()
{
    return WBaseScreen().run();
}
