/**
  *  \file client/scr-ship.cc
  *  \brief Starship Control Screen
  *
  *  This contains the ship screen tile widgets, and the main driver
  *  code.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "client/dlg-cargo.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "client/dlg-mission.h"
#include "client/marks.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/fcode.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/ship.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dlg-warp.h"
#include "game/mission.h"
#include "game/config.h"
#include "resmgr/resid.h"

namespace {
    /// Starship selection window.
    class WShipSelect : public WObjectSelectionTile {
        GUniverse& univ;
     public:
        WShipSelect(GUniverse& univ);
        ~WShipSelect();
        void drawData(GfxCanvas& can);
    };

    /// Ship scanner (with waypoint etc).
    class WShipScannerChartWidget : public WScannerChartWidget {
     public:
        WShipScannerChartWidget();
        ~WShipScannerChartWidget();
        void drawScanner(GfxCanvas& can);
        void update(int);
        void userMoveScannerTo(GPoint pt);
    };

    /// Starship screen.
    class WShipScreen : public WControlScreen {
        WShipScannerChartWidget scanner_chart;
        SignalConnection sig_conn, sig_ch_conn;
     public:
        WShipScreen();
        ~WShipScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void update();
        void setWaypointToScanner();
    };

    /// Ship equipment tile (crew/arms/damage).
    class WShipEquipmentTile : public WTile {
     public:
        WShipEquipmentTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
    };

    /// Ship cargo tile (ore/clans/cash).
    class WShipCargoTile : public WTile {
        void drawCargoBar(GfxCanvas& can, int x, int y, int wi, int have, int max, int color, const char* fmt);
     public:
        WShipCargoTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
        void addControls(UIGroup& g, ObjectHolder& h);
    };

    /// Ship mission tile (mission/PE/FC).
    class WShipMissionTile : public WTile {
     public:
        WShipMissionTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
        void addControls(UIGroup& g, ObjectHolder& h);
    };

    /// Ship movement tile (waypoint/speed/mass/ETA).
    class WShipMovementTile : public WTile {
     public:
        WShipMovementTile();
        void drawFixed(GfxCanvas& can);
        void drawData(GfxCanvas& can);
        void addControls(UIGroup& g, ObjectHolder& h);
    };
}

static int acquireWaypointPerm(const string_t& reason, GShip& ship);

/*************************** WShipSelect **************************/

WShipSelect::WShipSelect(GUniverse& univ)
    : WObjectSelectionTile(univ.getShipType(), univ.sig_ship_changed),
      univ(univ)
{ }

WShipSelect::~WShipSelect()
{ }

void
WShipSelect::drawData(GfxCanvas& can)
{
    GShip& ps = univ.getShip(getCurrent());

    // FIXME: this is very incomplet
    int y = 0, rem = 0;
    if (ps.isMarked()) {
        drawBackground(can, GfxRect(0, 0, 15, font_heights[FONT_NORMAL]));
        getSkin().getParameters().color = standard_colors[COLOR_GREENBLACK];
        drawSelection(can, getSkin(), extent.x + 5, extent.y + 7, 1, 2);
        rem = 15;
    }
    setColor(UIColor::tc_Static);
    outTextF(can, FONT_NORMAL, rem, y, extent.w - rem,
             ps.getFullName() + ", " + ps.getHull().getName());
    y += font_heights[FONT_NORMAL] * 3/2;
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             _("Location: ") + univ.getLocationName(ps.getPos(), locs_WW | locs_Orbit | locs_Verbose));
    y += font_heights[FONT_NORMAL];
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             _("Waypoint: ") + univ.getLocationName(ps.getWaypoint(), locs_WW | locs_Verbose));
    y += font_heights[FONT_NORMAL];
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             format(_("Distance: %.2f ly"), distFromDX(ps.getWaypointDX(),
                                                       ps.getWaypointDY())));
}

int
chooseShip()
{
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    int nships = univ.getShipType().getCount();
    if (nships == 0) {
        messageBox(_("You do not have any ships. "
                     "To build ships, use [B] on a starbase."),
                   _("Starships"));
        return 0;
    } else if (nships == 1) {
        return univ.getShipType().getCurrent();
    } else {
        return WShipSelect(univ).doStandardDialog(_("Select Ship"));
    }
}

/*************************** WShipEquipmentTile **************************/

WShipEquipmentTile::WShipEquipmentTile()
    : WTile(315, 6*font_heights[FONT_NORMAL]+3)
{ }

static void
drawShipEquipmentTile(UIAbstractTableWriter& tw)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));

    tw.newLColumn();

    const GShip& ship = getDisplayedTurn().getShipType().getShip();
    const GHull& hull = ship.getHull();

    /* engines */
    {
        string_t en = getEngine(ship.getEngine()).getName();
        if (ship.canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (hull.getNumEngines() > 1)
            en = format("%d x %s", hull.getNumEngines(), en);
        tw.writeVariableCell(UIColor::tc_Green, en);
    }

    /* primary */
    if (ship.getBeamCount() && ship.getBeamType()) {
        string_t bm = getBeam(ship.getBeamType()).getName();
        if (ship.getBeamCount() != 1)
            bm = format("%d x %s", ship.getBeamCount(), bm);
        tw.writeVariableCell(UIColor::tc_Green, bm);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* secondary */
    if (ship.getTorpType() && ship.getTorpLauncherCount()) {
        string_t tl = format(_("%d x %s launcher%!1{s%}"),
                             ship.getTorpLauncherCount(),
                             getTorp(ship.getTorpType()).getName());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d torpedo%!1{es%} aboard"), ship.getAmmo()))
                       : string_t(_("no torpedoes aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else if (ship.getBayCount()) {
        string_t tl = format(_("%d fighter bay%!1{s%}"), ship.getBayCount());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d fighter%!1{s%} ready"), ship.getAmmo()))
                       : string_t(_("no fighters aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    /* damage */
    if (ship.getDamage()) {
        int i = ship.getSupplies() / 5;
        int j = ship.getDamage() - i;
        UISkin::Color c = UIColor::tc_Red;
        if (j < 0) {
            j = 0;
            c = UIColor::tc_Yellow;
        }
        if (i)
            tw.writeVariableCell(c, format("%d%% (-%d: %d%%)", ship.getDamage(), i, j));
        else
            tw.writeVariableCell(c, format("%d%%", ship.getDamage()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* crew */
    if (ship.getCrew() != hull.getCrew()) {
        tw.writeVariableCell(ship.getCrew() > hull.getCrew() ? UIColor::tc_Yellow : UIColor::tc_Red,
                             format(_("%d of %d"), ship.getCrew(), hull.getCrew()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, numToString(ship.getCrew()));
    }
    tw.finish();
}

void
WShipEquipmentTile::drawFixed(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL,
                     UITableWriter::fl_IgnoreVariable);
    drawShipEquipmentTile(tw);
}

void
WShipEquipmentTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL,
                     UITableWriter::fl_IgnoreFixed | UITableWriter::fl_Fill);
    drawShipEquipmentTile(tw);
}

/***************************** WShipCargoTile ****************************/

WShipCargoTile::WShipCargoTile()
    // FIXME: magic numbers
    : WTile(315, 4 * font_heights[FONT_NORMAL] + 14 + 3)
{ }

static void
drawShipCargoTile(UIAbstractTableWriter& tw)
{
    // Column 1: ore names
    // Column 2: ore amounts
    // Column 3: space
    // Column 4: rest
    // Column 5: rest amounts
    // Column 6: units

    GShip& s = getDisplayedTurn().getShipType().getShip();

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, "Neu:");
    tw.writeFixedCell(UIColor::tc_Static, "Tri:");
    tw.writeFixedCell(UIColor::tc_Static, "Dur:");
    tw.writeFixedCell(UIColor::tc_Static, "Mol:");

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9999 kt"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s.getCargo(el_Neutronium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s.getCargo(el_Tritanium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s.getCargo(el_Duranium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s.getCargo(el_Molybdenum)));

    tw.newLColumn(5);           // spacer

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Colonists:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Total Mass:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999"));
    tw.writeVariableCell(UIColor::tc_Green, clansToString(s.getCargo(el_Colonists)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s.getCargo(el_Supplies)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s.getCargo(el_Money)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s.getMass()));

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Green, "");
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, "mc");
    tw.writeFixedCell(UIColor::tc_Green, "kt");

    tw.finish();
}

void
WShipCargoTile::drawFixed(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_IgnoreVariable);
    drawShipCargoTile(tw);
}

void
WShipCargoTile::drawData(GfxCanvas& can)
{
    GfxRect pos = getInnerSize(0, 0);
    GShip& s = getDisplayedTurn().getShipType().getShip();
    const int y = font_heights[FONT_NORMAL];

    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_IgnoreFixed);
    drawShipCargoTile(tw);

    int cargo = s.getHull().getCargo();
    drawCargoBar(can, pos.x + 5,   pos.y + 4*y + 4, 130, s.getCargo(el_Neutronium), s.getHull().getFuel(), COLOR_GREENBLACK, _("Fuel: %d free"));
    drawCargoBar(can, pos.x + 145, pos.y + 4*y + 4, 130, cargo - s.getFreeCargo(), cargo, COLOR_BLUE, _("Cargo: %d free"));
}

void
WShipCargoTile::drawCargoBar(GfxCanvas& can, int x, int y, int wi, int have, int max,
                             int color, const char* fmt)
{
    drawFrameDown(can, GfxRect(x, y, wi, 10));
    ++x;
    ++y;
    wi -= 2;

    int split = max ? wi * have / max : 0;
    if (have > 0 && split == 0)
        split = 1;
    if (have >= max)
        split = wi;
    else if (split >= wi)
        --split;

    if (split)
        drawSolidBar(can, GfxRect(x, y, split, 8), standard_colors[color]);
    if (split != wi)
        drawSolidBar(can, GfxRect(x + split, y, wi-split, 8), standard_colors[COLOR_BLACK]);

    getSkin().getParameters().just_x = 1;
    getSkin().getParameters().just_y = 0;
    if (have > max)
        getSkin().getParameters().color = standard_colors[COLOR_RED];
    else
        getSkin().getParameters().color = standard_colors[COLOR_WHITE];
    getSkin().getParameters().bgcolor = GFX_TRANSPARENT;
    getSkin().getParameters().font = fonts[FONT_SMALL];
    ::outTextF(can, getSkin(), x + wi/2, y-1, wi, format(fmt, max-have));
}

void
WShipCargoTile::addControls(UIGroup& g, ObjectHolder& h)
{
    addControl(g, GfxRect(295, 0, 20, 20), h.add(new UIButton("C", 'c')));
    addControl(g, GfxRect(270, 0, 20, 20), h.add(new UIButton("U", 'u')));
}

/**************************** WShipMissionTile ***************************/

WShipMissionTile::WShipMissionTile()
    : WTile(315, 9*font_heights[FONT_NORMAL]/2)
{ }

void
WShipMissionTile::drawFixed(GfxCanvas& can)
{
    const int y = font_heights[FONT_NORMAL];
    setColor(UIColor::tc_Static);
    outText(can, FONT_NORMAL, 5,   0, _("Mission:"));
    outText(can, FONT_NORMAL, 5,   y, _("Primary Enemy:"));
    outText(can, FONT_NORMAL, 5, 2*y, _("Friendly Code:"));
}

void
WShipMissionTile::drawData(GfxCanvas& can)
{
    const int y = font_heights[FONT_NORMAL];
    const int x = 120;

    // cannot be const.
    GShip& ship = getDisplayedTurn().getShipType().getShip();

    const GMission* msn =
        GMissionList::getInstance().getMissionByNumber(ship.getMission(),
                                                       GPlayerSet(raceId(ship.getRealOwner())));

    setColor(UIColor::tc_Green);
    if (msn)
        outTextF(can, FONT_NORMAL, x-50, 0, 214,
                 msn->getName(ship));
    else
        outTextF(can, FONT_NORMAL, x-50, 0, 214,
                 format("M%d I%d T%d", ship.getMission(), ship.getIntId(), ship.getTowId()));

    outTextF(can, FONT_NORMAL, x,    y, 164-25,
             (ship.getPE()
              ? getShortName(ship.getPE())
              : string_t(_("none"))));
    outTextF(can, FONT_NORMAL, x,  2*y, 100, ship.getFCode());

    // FIXME
//     string_t comment = "FIXME";
//     if (comment.empty()) {
//         setColor(UIColor::tc_Static);
//         comment = _("(press [F9] to add a note)");
//     } else {
//         setColor(UIColor::tc_Yellow);
//     }
//     outTextF(can, FONT_NORMAL, 5,  7*y/2, 265, comment);
}

void
WShipMissionTile::addControls(UIGroup& g, ObjectHolder& h)
{
    addControl(g, GfxRect(295, 0, 20, 20),  h.add(new UIButton("M", 'm')));
    addControl(g, GfxRect(295, 25, 20, 20), h.add(new UIButton("E", 'e')));
    addControl(g, GfxRect(295, 50, 20, 20), h.add(new UIButton("F", 'f')));
//     addControl(g, GfxRect(270, 50, 20, 20), h.add(new UIButton("B", 'b')));
//     addControl(g, GfxRect(285, 75, 30, 20),  h.add(new UIButton("F9", SDLK_F9)));
}

/*************************** WShipMovementTile ***************************/

WShipMovementTile::WShipMovementTile()
    : WTile(GfxPoint(315, 7*font_heights[FONT_NORMAL]))
{ }

static void
drawShipMovementTile(UIAbstractTableWriter& tw, bool fixed_only)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Location:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Waypoint:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Distance:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Warp Factor:"));
    tw.writeFixedCell(UIColor::tc_Static, _("E.T.A.:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fuel Usage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Engine Load:"));

    if (fixed_only)
        return;                 // save time by not computing movement info

    /* compute stuff */
    GShip& s = getDisplayedTurn().getShipType().getShip();
    GShipTurnPredictor crystal_ball(getDisplayedTurn().getCurrentUniverse(), s.getId());
    crystal_ball.computeMovement();

    const bool is_hyper = s.isHyperdriving();

    tw.newLColumn();

    /* location */
    tw.writeVariableCell(UIColor::tc_Green, getDisplayedTurn().getCurrentUniverse().getLocationName(s.getPos(), locs_WW | locs_Verbose));

    /* waypoint */
    if (s.getMission() == GMission::msn_Intercept
        && s.getIntId() > 0
        && s.getIntId() <= NUM_SHIPS
        && getDisplayedTurn().getShip(s.getIntId()).isExistant())
    {
        tw.writeVariableCell(UIColor::tc_Green,
                             getDisplayedTurn().getShip(s.getIntId()).getName());
    } else {
        tw.writeVariableCell(UIColor::tc_Green,
                             getDisplayedTurn().getCurrentUniverse().getLocationName(s.getWaypoint(),
                                                                locs_WW | locs_Verbose | locs_Ships));
    }

    /* distance */
    double dist = distFromDX(s.getWaypointDX(), s.getWaypointDY());
    tw.writeVariableCell(!is_hyper
                         ? UIColor::tc_Green
                         : dist < 20
                         ? UIColor::tc_Red
                         : dist < 340 || dist > 360  // FIXME
                         ? UIColor::tc_Yellow
                         : UIColor::tc_Green,
                         format(_("%.2f ly"), dist));

    /* warp */
    if (s.getWarp() == 0 && (s.getWaypointDX() | s.getWaypointDY())) {
        tw.writeVariableCell(UIColor::tc_Red, _("not moving"));
    } else {
        // Borland sucks.
        string_t st;
        if (is_hyper)
            st = _("Hyperdrive");
        else if (s.getWarp() == 0)
            st = _("not moving");
        else
            st = format(_("Warp %d"), s.getWarp());
        tw.writeVariableCell(UIColor::tc_Green, st);
    }

    /* e.t.a. */
    {
        string_t eta;
        UISkin::Color c = UIColor::tc_Green;

        if (s.getWaypointDX() == 0 && s.getWaypointDY() == 0) {
            eta = _("at waypoint");
        } else if (is_hyper) {
            eta = _("unknown"); // FIXME?
        } else if (s.getWarp() == 0) {
            c = UIColor::tc_Red;
            eta = _("not moving");
        } else if (crystal_ball.getTurnCount() >= crystal_ball.MOVEMENT_TIME_LIMIT) {
            eta = _("too long");
        } else {
            eta = format(_("%d turn%!1{s%}"), crystal_ball.getTurnCount());
        }
        tw.writeVariableCell(c, eta);
    }

    /* fuel usage */
    {
        string_t fuel;
        UISkin::Color c = UIColor::tc_Green;

        fuel = format(_("%d kt"), crystal_ball.getFuelUsed());
        if (crystal_ball.getTurnCount() >= crystal_ball.MOVEMENT_TIME_LIMIT)
            c = UIColor::tc_Yellow;
        else if (crystal_ball.getFuelUsed() > s.getCargo(el_Neutronium))
            c = UIColor::tc_Red;

        // FIXME: cloak fuel
        tw.writeVariableCell(c, fuel);
    }

    /* load */
    int towee_mass = 0;
    if (s.getMission() == GMission::msn_Tow
        && s.getTowId() > 0 && s.getTowId() <= NUM_SHIPS
        && getDisplayedTurn().getShip(s.getTowId()).isExistant())
    {
        towee_mass = getDisplayedTurn().getShip(s.getTowId()).getMass();
    }
    tw.writeVariableCell(UIColor::tc_Green,
                         format(_("%d kt"), s.getMass() + towee_mass));

    tw.finish();
}

void
WShipMovementTile::drawFixed(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 20), FONT_NORMAL,
                     UITableWriter::fl_IgnoreVariable);
    drawShipMovementTile(tw, true);
}

void
WShipMovementTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 20), FONT_NORMAL,
                     UITableWriter::fl_IgnoreFixed | UITableWriter::fl_Fill);
    drawShipMovementTile(tw, false);
}

void
WShipMovementTile::addControls(UIGroup& g, ObjectHolder& h)
{
    GfxRect r = getInnerSize(0, 0);  // we want to know the height and width
    addControl(g, GfxRect(r.w - 20, r.h - 20, 20, 20), h.add(new UIButton("A", 'a')));
    addControl(g, GfxRect(r.w - 20, r.h - 45, 20, 20), h.add(new UIButton("W", 'w')));
}

/************************ WShipScannerChartWidget ************************/

WShipScannerChartWidget::WShipScannerChartWidget()
{ }

WShipScannerChartWidget::~WShipScannerChartWidget()
{ }

void
WShipScannerChartWidget::drawScanner(GfxCanvas& can)
{
    const GChartViewport& vp = getViewport();

    GPoint                wp = getDisplayedTurn().getShipType().getShip().getWaypoint();
    GfxPoint       wp_screen = vp.scaleXY(wp);

    if (wp != getScannerPosition() && wp_screen != vp.getCenter()) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_RED];

        drawLine(can, parm, vp.getCenter().x, vp.getCenter().y, wp_screen.x, wp_screen.y);
        drawEndBig(can, parm, wp_screen);
    }

    WScannerChartWidget::drawScanner(can);
}

void
WShipScannerChartWidget::update(int)
{
    drawWidget(false);
}

// void
// WShipScannerChartWidget::handleScannerChange(bool dblclick)
// {
//     if (dblclick) {
//         GGameTurn& trn = getDisplayedTurn();
//         if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
//             trn.getShip(id).setWaypoint(getScannerPosition());
//             trn.getCurrentUniverse().doUpdates();
//         }
//     }
// }

void
WShipScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    bool scanner_tied_to_waypoint = false;
    if (scanner_tied_to_waypoint) {
        // FIXME
        setScannerPosition(pt);
    } else {
        setScannerPosition(pt);
    }
}

/****************************** WShipScreen ******************************/

WShipScreen::WShipScreen()
    : WControlScreen(getDisplayedTurn().getShipType(), 1, GShip::ss_DirtyHead),
      scanner_chart()
{
    addTile(holder.add(new WControlScreenHeaderTile(*this, WControlScreenHeaderTile::bName + WControlScreenHeaderTile::bAuto + WControlScreenHeaderTile::bCycle + WControlScreenHeaderTile::bAdd,
                                                    GShip::ss_DirtyHead)));
    addTile(holder.add(new WShipEquipmentTile()),    _("Equipment & Crew:"));
    addTile(holder.add(new WShipCargoTile()),        _("Aboard:"));
    addTile(holder.add(new WShipMissionTile()),      _("Mission:"));
    addTile(holder.add(new WShipMovementTile()),     _("Movement:"));
    addChart(scanner_chart);
    sig_conn = getDisplayedTurn().getCurrentUniverse().
        sig_ship_changed.add(static_cast<WControlScreen*>(this),
                             &WControlScreen::updateId);
    sig_ch_conn = sig_changed.add(static_cast<WShipScannerChartWidget*>(&scanner_chart),
                                  &WShipScannerChartWidget::update);

    scanner_chart.sig_dblclick.add(this, &WShipScreen::setWaypointToScanner);
}

WShipScreen::~WShipScreen()
{ }

void
WShipScreen::update()
{
    int sid = getType().getCurrent();
    const GShip& ship = getType().getUniverse().getShip(sid);
    setData(ResId(res::ship, ship.getHull().getPicture(), ship.getHullId()),
            ship.getPos(),
            ship.getName(),
            format(_("(Id #%d, %s)"), sid, ship.getHull().getName()));
    scanner_chart.setScannerPosition(scanner_chart.clipScannerPosition(ship.getWaypoint()));
}

static void
doShipXchg(WControlScreen& scr)
{
    WScanner& scan             = scr.getScanner();
    WScannerChartWidget* chart = scan.getChart();
    int n;
    if (scan.getScannedShipId()) {
        n = scr.getType().getUniverse().getNumberOfPlayedShipsAt(scr.getScannerPosition());
        if (chart->getCenterPosition() == chart->getScannerPosition())
            --n;                // discount current ship
    } else {
        n = 0;
    }

    if (n)
        putKey(ss_Ctrl + SDLK_F1);
    else
        putKey(ss_Ctrl + SDLK_F2);
}

/** Attempt to acquire permission to change waypoint. (this is used for fleets)
    \param reason reason, for user
    \param ship   ship we want to change
    \returns 0 if permission not granted, ship id to change waypoint on
    specified ship */
static int
acquireWaypointPerm(const string_t& reason, GShip& ship)
{
    if (!ship.getWaypointPermission()) {
        // FIXME: display dialog, leave fleet, ...
        (void) reason;
        return 0;
    } else {
        return ship.getId();
    }
}

void
doShipTransferFor(GShip& ship, GObjectId chosen)
{
    GMultiTransaction mt;
    GShipTransfer&   left = mt.addNewSubtransaction(new GShipTransfer(ship));
    GCargoContainer* right;

    if (!chosen.isGameObject()) {
        /* jettison */
        right = &mt.addNewSubtransaction(new GShipTransporter(ship));
    } else if (GShip* s = dynamic_cast<GShip*>(&chosen.getObject())) {
        if (s->getOwner() == ship.getOwner()) {
            /* ship <-> ship */
            ASSERT(s->haveData());
            right = &mt.addNewSubtransaction(new GShipTransfer(*s));
        } else {
            /* ship <-> foreign ship */
            if (ship.getCargoHandler().isFacilityInUse(GShip::sf_Transfer)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                left.doCancel();
                return;
            }
            if (ship.getTransporter(GShip::t_Transfer).id != 0
                && ship.getTransporter(GShip::t_Transfer).id != s->getId()
                && ship.isTransporterActive(GShip::t_Transfer))
            {
                messageBox(format(_("The cargo transporter is already transferring cargo "
                                    "to ship #%d."),
                                  ship.getTransporter(GShip::t_Transfer).id), _("Cargo Transfer"));
                // FIXME: state name of current target ship here
                // FIXME: we could cancel the transfer here
                return;
            }
            // ship.cancelTranporter(true);
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Transfer, *s));
        }
    } else if (GPlanet* p = dynamic_cast<GPlanet*>(&chosen.getObject())) {
        if (p->getOwner() == ship.getOwner() && p->havePlanetData()) {
            right = &mt.addNewSubtransaction(new GPlanetTransfer(*p));
        } else {
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Unload, *p));
        }
    } else {
        FAIL();
        right = 0;
    }
    
    doCargoTransfer(mt, mt.addNewSubtransaction(new GCargoTransfer(left, *right)));
}

static void
doShipCargoTransfer(GUniverse& univ, GShip& ship)
{
    GObjectList list;
    int pid = univ.getPlanetAt(ship.getPos());
    if (pid)
        list.addObject(univ.getPlanetType(), univ.getPlanet(pid));
    else
        list.addDivider(_("Jettison into space"), -1);

    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (i != ship.getId() && univ.getShip(i).isExistant() && univ.getShip(i).getPos() == ship.getPos())
            list.addObject(univ.getShipType(), univ.getShip(i));

    GObjectId chosen = WObjectSelectionList(0, 15, 320, list).doStandardDialog(_("Transfer cargo to..."), 0 /* FIXME */);
    if (chosen.isValid())
        doShipTransferFor(ship, chosen);
}

static void
doShipNavigationChart(GUniverse& univ, GShip& ship)
{
    int size = ui_root->getExtent().h;

    UIWindow the_window(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    the_window.setLayout(UIVBoxLayout::instance5);

    WScannerChartWidget chart;
    chart.setCenterPosition(ship.getPos());
    chart.setScannerPosition(chart.clipScannerPosition(ship.getWaypoint()));
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    the_window.add(chart);

    UIGroup buttons(UIHBoxLayout::instance5);
    ObjectHolder h;
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10, UIButton::aStop)));
    buttons.add(h.add(new UIButton(cm_Escape, "ESC",        SDLK_ESCAPE, UIButton::aStop)));
    the_window.add(buttons);
    the_window.setExtent(GfxRect(0, 0, size - 50, size - 10));
    the_window.doLayout();

    ui_root->centerWidget(the_window);
    ui_root->add(the_window);
    int rv = ui_root->run();
    ui_root->close(the_window);

    if (rv == cm_Enter)
        ship.setWaypoint(chart.getScannerPosition());
}

void
WShipScreen::setWaypointToScanner()
{
    GGameTurn& trn = getDisplayedTurn();
    if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
        trn.getShip(id).setWaypoint(getScannerPosition());
        trn.getCurrentUniverse().doUpdates();
    }
}

static void
doChangePE(GUniverse& univ, GShip& ship)
{
    UIStringList d;
    d.add(0, _("0 - none"));
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        d.add(i, format("%X - %s", i, getShortName(i)));
    
    UIStandardListbox lb(0, -1, 240, d);
    lb.scrollToKey(ship.getPE());
    if (lb.doStandardListbox(_("Change Primary Enemy"), 0))
        ship.setPE(lb.getKey());
    univ.doUpdates();
}

bool
WShipScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        GUniverse& univ = getType().getUniverse();
        GShip&     ship = univ.getShipType().getShip();
        switch (e.key.code) {
         case 'n': {
             string_t s = ship.getName();
             if (UIInputLine::run(s, 20, format(_("Rename Starship #%d"),
                                                ship.getId()),
                             _("Enter new name:"), 20, 0))
                 ship.setName(s);
             univ.doUpdates();
             return true;
         }
         case 'm':
            doShipMissionChange(univ, ship);
            return true;
         case 'e':
            doChangePE(univ, ship);
            return true;
         case 'a':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                doShipNavigationChart(univ, univ.getShip(id));
                univ.doUpdates();
            }
            return true;
         case 'c':
            doShipCargoTransfer(univ, ship);
            univ.doUpdates();
            return true;
         case ss_Ctrl + 'j':
            if (univ.getPlanetAt(ship.getPos()) == 0) {
                doShipTransferFor(ship, GObjectId());
                univ.doUpdates();
            }
            return true;
         case ss_Ctrl + 's':
            if (ship.getTransporter(GShip::t_Transfer).id) {
                doShipTransferFor(ship, GObjectId(univ.getShipType(),
                                                  univ.getShip(ship.getTransporter(GShip::t_Transfer).id)));
                univ.doUpdates();
            }
            return true;
         case ss_Ctrl + 'p':
            // FIXME: should possibly also handle beam-up orders
            if (ship.getTransporter(GShip::t_Unload).id) {
                doShipTransferFor(ship, GObjectId(univ.getPlanetType(),
                                                  univ.getPlanet(ship.getOrbit())));
                univ.doUpdates();
            }
            return true;
         case 'u':
            if (ship.getOrbit()) {
                doShipTransferFor(ship, GObjectId(univ.getPlanetType(),
                                                  univ.getPlanet(ship.getOrbit())));
                univ.doUpdates();
            } else {
                messageBox(_("We are not orbiting a planet, sir. "
                             "If you want to jettison cargo, use the "
                             "'c' command instead."),
                           _("Cargo Transfer"));
            }
            return true;
         case 'x':
            doShipXchg(*this);
            return true;
         case 'f': {
             GFCodeList list(GFCodeList::getInstance(), ship);
             WFCodeWindow dialog(_("Change Friendly Code"), list);
             dialog.setFCode(ship.getFCode());
             ui_root->centerWidget(dialog);
             ui_root->add(dialog);
             int rv = ui_root->run();
             ui_root->close(dialog);
             if (rv == cm_Enter) {
                 ship.setFCode(dialog.getFCode());
                 univ.doUpdates();
             }
             return true;
         }
         case 'w':
            if (int id = acquireWaypointPerm(_("speed"), ship)) {
                (WShipSpeedSelector(univ.getShip(id)).doStandardDialog());
            }
            return true;
         case ss_Ctrl + 'w':
            setWaypointToScanner();
            return true;
         case ss_Ctrl + 'i':
            if (ship.getIntId())
                putCommand(cm_CS_GotoCS, 1, ship.getIntId());
            return true;
         case ss_Ctrl + 't':
            if (ship.getTowId())
                putCommand(cm_CS_GotoCS, 1, ship.getTowId());
            return true;
         case SDLK_F8:
         case ss_Shift + SDLK_F8:
            putCommand(cm_CS_GotoCS, 3, ship.getOrbit());
            return true;
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Shift + ss_Ctrl + SDLK_TAB: {
             GObjectList l;
             l.addObjectsAt(univ, pos, 0, 0);
             l.postprocessUser(false);
             int n = l.getIndexFor(GObjectId(univ.getShipType(), ship));
             int overflow = 0;
             ASSERT(n >= 0);
             do {
                 if ((e.key.code & ss_Shift) == 0) {
                     if (++n >= l.getNumGameObjects())
                         n = 0, ++overflow;
                 } else {
                     if (--n < 0)
                         n = l.getNumGameObjects()-1, ++overflow;
                 }
             } while (!l[n].getType().getObject(l[n].getId()).isMarked() && (e.key.code & ss_Ctrl) && overflow < 2);
             if (overflow < 2)
                 putCommand(cm_CS_GotoCS, getScreenId(), l[n].getId());
             break;
         }
            /* ship screen keys:

               m-i, m-t         // search xceptor/tower
               c-n              // list next turn
               q                // toggle independantmove
               b                // base commands
               ins              // sim
               s                // spec
               r                // remote
               g                // give
               d                // distribute cargo
               f9               // comment
               f10              // fleet
               s-f10            // join fleet/goto fleet
               m-r              // reset x/y
            */
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

int
doShipScreen()
{
    return WShipScreen().run();
}
