/**
  *  \file global.h
  *  \brief Global Turn Data
  *
  *  This module provides classes to manage data of a complete turn,
  *  i.e. which contain information about all objects at once.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GLOBAL_H
#define PCC_V2_GLOBAL_H

#include <map>
#include "io/streams.h"
#include "io/dirs.h"
#include "game/vcrdb.h"
#include "game/msg.h"
#include "game/gen.h"
#include "game/mine.h"
#include "game/cmds.h"
#include "game/gamedirs.h"      // re-exported
#include "cpluslib/sigslot.h"
#include "cpluslib/assert.h"
#include "game/ion.h"
#include "game/player.h"
#include "game/univ.h"

/**************************** Data Structures ****************************/

//! Chart Bounding Box.
struct GChartBBox {
    enum {
        MAX_ZOOM = 10
    };

    GPoint min, max;
    int    max_zoom;
    int    min_zoom;

    void initFromUniverse(GUniverse& trn);
    void initMinZoom();
    void initMaxZoom();

    void addPoint(GPoint pt);
    void addCircle(GPoint pt, int radius);
};

//! Everything that represents a game turn.
class GGameTurn {
    /* I'm still unsure what to do with inbox and VCRs */
    struct Player {
        GOutbox           outbox;
        GGen              gen;
        GCommandContainer commands;
        Player(int pid);
        ~Player();
    };
 private:
    GUniverse         dat, dis;

    GInbox            inbox;
    Ptr<GVcrDatabase> vcrs;
    GChartBBox        chart_bbox;
    Player*           player_data[NUM_PLAYERS];
    GPlayerSet        playing_set, data_set;

    Player& createPlayerData(int player);
    const Player& getPlayerDataNC(int player) const;
    Player& getPlayerData(int player);

    void loadPlanetNames();
    void loadStormNames();
public:
    GGameTurn();
    ~GGameTurn();

    void init();
    void done();

    GUniverse& getCurrentUniverse()
        { return dat; }
    GUniverse& getPreviousUniverse()
        { return dis; }

    /// \name Player Status Accessors
    //@{
    void setGen(int player, const GGen& gen);
    GGen& getGen(int player)
        { return getPlayerData(player).gen; }
    const GGen& getGen(int player) const
        { return getPlayerDataNC(player).gen; }

    int getTurnNumber() const
        { return dat.getTurnNumber(); }
    
    bool isPlaying(int player_nr) const
        { return playing_set.contains(player_nr); }
    bool isData(int player_nr) const
        { return data_set.contains(player_nr); }
    void setHaveData(int player_nr);
    //@}

    /// \name Object Accessors
    //@{
    GShip& getShip(int sid)
        { return dat.getShip(sid); }
    const GShip& getShip(int sid) const
        { return dat.getShip(sid); }
    GPlanet& getPlanet(int pid)
        { return dat.getPlanet(pid); }
    const GPlanet& getPlanet(int pid) const
        { return dat.getPlanet(pid); }
    GIonStorm& getIonStorm(int sid)
        { return dat.getIonStorm(sid); }
    const GIonStorm& getIonStorm(int sid) const
        { return dat.getIonStorm(sid); }
    GMinefield* getMinefield(int id, int cause)
        { return dat.getMinefield(id, cause); }
    GMinefield* getMinefield(int id)
        { return getMinefield(id, -1); }
    GUniverse::MinefieldMap::iterator getMinefieldBegin()
        { return dat.getMinefieldBegin(); }
    GUniverse::MinefieldMap::iterator getMinefieldEnd()
        { return dat.getMinefieldEnd(); }
    //@}

    /// \name Locations
    //@{
    GChartBBox& getChartBBox()
        { return chart_bbox; }
    //@}

    /// \name Message Accessors
    //@{
    GInbox& getInbox()
        { return inbox; }
    GInbox const& getInbox() const
        { return inbox; }
    GOutbox const& getOutbox(int player) const
        { return getPlayerDataNC(player).outbox; }
    GOutbox& getOutbox(int player)
        { return getPlayerData(player).outbox; }
    GVcrDatabase& getVcrs() const
        { return *vcrs; }
    void setVcrs(Ptr<GVcrDatabase> db)
        { vcrs = db; }
    GCommandContainer& getCommands(int player)
        { return getPlayerData(player).commands; }
    //@}

    /// \name Meta-Information
    //@{
    GShipType& getShipType()
        { return dat.getShipType(); }
    GPlanetType& getPlanetType()
        { return dat.getPlanetType(); }
    GBaseType& getBaseType()
        { return dat.getBaseType(); }
    GIonStormType& getIonStormType()
        { return dat.getIonStormType(); }
    //@}

    /// \name Loading
    //@{
    void loadPlanetXY(Stream&);
    void postprocess();
    //@}
};

/******************************* Accessors *******************************/

// These are exposed here only for the following inline functions
extern Ptr<GGameTurn> current_turn_internal;
extern Ptr<GGameTurn> display_turn_internal;

/** Get global turn instance. */
inline GGameTurn&  getCurrentTurn()   { return *current_turn_internal; }
inline GGameTurn&  getDisplayedTurn() { return *display_turn_internal; }

void setDisplayedTurn(Ptr<GGameTurn> disp);

extern Signal0 sig_turn_changed;

/** Global Data. */
void initTurn();
void initGameGlobals(string_t gamedir, string_t rootdir);

#endif
