/**
  *  \file game/ion.cc
  *  \brief Ion Storms
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ion.h"
#include "util/nls.h"
#include "game/global.h"
#include "cpluslib/format.h"
#include "game/utildata.h"

/** Create ion storm. Initializes the object for a non-existant
    storm. */
GIonStorm::GIonStorm(int id)
    : id(id)
{
    data.volt = 0;              // non-existant
}

GIonStorm::~GIonStorm()
{ }

/** Add data from koreX.dat. */
void
GIonStorm::addKoreData(const TKoreStorm& kstorm)
{
    data = kstorm;
    grow = (kstorm.volt & 1) != 0;
}

/** Add data from utilX.dat.
    \pre ustorm.storm_id == getId() */
void
GIonStorm::addUtilData(const TUtil17Ion& ustorm)
{
    ASSERT(ustorm.storm_id == id);
    data.x       = ustorm.x;
    data.y       = ustorm.y;
    data.volt    = ustorm.volt;
    data.heading = ustorm.heading;
    data.speed   = ustorm.speed;
    data.radius  = ustorm.radius;
    grow         = ustorm.grows != 0;
    // FIXME: ustorm.storm_class is ignored.
}

string_t
GIonStorm::getName() const
{
    return name;
}

string_t
GIonStorm::getFullName() const
{
    return format(_("%s (Ion storm #%d)"), name, id);
}

int
GIonStorm::getId() const
{
    return id;
}

int
GIonStorm::getOwner() const
{
    return 0;
}

bool
GIonStorm::isExistant() const
{
    return data.volt != 0;
}

GPoint
GIonStorm::getPos() const
{
    return GPoint(data.x, data.y);
}

int
GIonStorm::getRadius() const
{
    return data.radius;
}

/***************************** GIonStormType *****************************/

GIonStormType::GIonStormType(GUniverse& univ)
    : GObjectType(NUM_STORMS, univ)
{ }

bool
GIonStormType::isExistant(int index)
{
    return index > 0 && index <= NUM_STORMS && getUniverse().getIonStorm(index).isExistant();
}

string_t
GIonStormType::getClassName()
{
    return string_t(_("Ion Storm"));
}

GIonStorm&
GIonStormType::getObject(int index)
{
    return getUniverse().getIonStorm(index);
}
