/**
  *  \file game/planet.h
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_PLANET_H
#define PCC_V2_GAME_PLANET_H

#include "game/struct.h"
#include "game/coord.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/cargo.h"
#include "game/player.h"

class GShip;
class GUniverse;
class GTurnfile;

/// Native races.
enum {
    nr_None,
    nr_Humanoid,
    nr_Bovinoid,
    nr_Reptilian,
    nr_Avian,
    nr_Amorphous,
    nr_Insectoid,
    nr_Amphibian,
    nr_Ghipsoldal,
    nr_Siliconoid
};

/// Shipyard actions. Value of the shipyard_action field.
enum {
    sya_None,                   ///< No operation.
    sya_Fix,                    ///< Fix a ship.
    sya_Recycle                 ///< Recycle a ship.
};

/// Planetary Structures.
enum TPlanetaryStructure {
    str_Mines,
    str_Factories,
    str_Defense,
    str_BaseDefense
};

class GPlanet : public GMapObject {
 private:
    GPlanet();
    GPlanet(const GPlanet&);
    GPlanet& operator=(const GPlanet&);

    static int getMaxStructuresFormula(int32 clans, int threshold);

    struct FullInfo {
        TPlanet data;
        int16   mineral_time, colonist_time, native_time, cash_time;
        bool    known_to_have_natives;

        FullInfo(int id, bool known);
    };
 public:
    enum {
        PLANET_EXISTS      = 4,   ///< This planet exists.
        PLANET_OWNORBIT    = 8,   ///< Own ship(s) in orbit.
        PLANET_ALLYORBIT   = 16,  ///< Allied ship(s) in orbit.
        PLANET_ENEMYORBIT  = 32,  ///< Enemy ship(s) in orbit.
        PLANET_GALLYORBIT  = 64,  ///< Guessed allied ship(s) in orbit.
        PLANET_GENEMYORBIT = 128  ///< Guessed enemy ship(s) in orbit.
    };
 private:
    int id;

    // Planet data
    FullInfo*  full_info;
    TStarbase* base_dat;
    int        x, y;
    bool       known_to_not_exist;
    string_t   name;

    // Internal flags
    int flags;

    // Source flags
    GPlayerSet psrc_flags, bsrc_flags;

    // Cargo reservation management
    GCargo cargo;

    FullInfo& createInfo(bool known);
    const TPlanet& getDisplayData() const;
 public:
    GPlanet(int id);
    ~GPlanet();

    // Loading and Saving
    void addPlanetData(const TPlanet& data, GPlayerSet srcf);
    void addBaseData(const TStarbase& data, GPlayerSet srcf);
    void setXY(int x, int y);
    void setName(const string_t& name);
    void setKnownToNotExist(bool value = true);

    /** Get a copy of the planet data in pdata format. */
    void getPlanetData(TPlanet& dat) const
        { dat = full_info->data; }
    /** Get a copy of the starbase data in bdata format. */
    void getBaseData(TStarbase& bdat) const
        { bdat = *this->base_dat; }
    /** True iff we have full information. */
    bool havePlanetData() const
        { return psrc_flags.nonempty(); }
    /** True iff we have starbase information. */
    bool haveBaseData() const
        { return base_dat != 0; }
    GPlayerSet getPlanetSrcFlags() const
        { return psrc_flags; }
    GPlayerSet getBaseSrcFlags() const
        { return bsrc_flags; }
    void applyTurnCommand(GTurnfile& trn, int index);
    void makePlanetTurnCommands(GTurnfile& trn, const GPlanet& dis) const;
    void makeBaseTurnCommands(GTurnfile& trn, const GPlanet& dis) const;

    /* GMapObject interface */
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;

    /* Accessors: Status */
    bool hasBase() const;
    int getStatus() const
        { return flags; }
    bool hasAnyInfo() const
        { return full_info != 0; }

    /* Accessors: Structures */
    mp16_t getStructures(TPlanetaryStructure ps) const;
    void   setStructures(TPlanetaryStructure ps, mp16_t count);
    int    getMaxStructures(TPlanetaryStructure ps) const
        { return getMaxStructures(ps, getColonists(), hasBase()); }
    static int getMaxStructures(TPlanetaryStructure ps, int32 clans, bool base);

    /* Accessors: Colonists */
    int    getColonistChange(int tax, int mifa) const;
    int    getColonistChange() const;

    mn16_t getColonistHappiness() const;
    void   setColonistHappiness(mn16_t happiness);
    mp16_t getColonistTax() const;
    void   setColonistTax(mp16_t tax);
    mn32_t getColonists() const;
    void   setColonists(mn32_t clans);

    bool isOwnerKnown() const
        { return mp16_t(getDisplayData().owner).isKnown(); }

    /* Formulas: colonists */
    int32 getColonistDue(int tax) const;
    int32 getColonistDueLimited(int tax, int32& rem_inc) const;
    int32 getMaxSupportedColonists(int player) const;
    int32 getMaxSupportedColonists() const
        { return getMaxSupportedColonists(getOwner()); }

    /* Accessors: Natives */
    int getNativeChange() const;
    int getNativeChange(int tax, int mifa) const;

    mp16_t getNativeGovernment() const;
    void   setNativeGovernment(mp16_t gov);
    mn16_t getNativeHappiness() const;
    void   setNativeHappiness(mn16_t happiness);
    mp16_t getNativeRace() const;
    void   setNativeRace(mp16_t race);
    mp16_t getNativeTax() const;
    void   setNativeTax(mp16_t tax);
    void   setNatives(mp32_t natives);
    mp32_t getNatives() const;

    bool isKnownToHaveNatives() const;
    void setKnownToHaveNatives(bool known);

    int32    getNativeDue(int tax) const;
    int32    getNativeDueLimited(int tax, int32 rem_inc) const;
    int32    getBovinoidSupplyContribution() const;
    int32    getBovinoidSupplyContributionLimited() const;

    /* Accessors: FCode */
    string_t getFCode() const;
    void     setFCode(const string_t& str);
    bool     isFCodeKnown() const;

    /* Accessors: Base building */
    bool     isBuildingBase() const
        { return full_info && full_info->data.build_base; }
    void     setBuildBaseFlag(bool b);

    /* Accessors: Environment */
    mp16_t getOreDensity(int index) const;
    void   setOreDensity(int index, mp16_t amount);
    mp32_t getOreGround(int index) const;
    void   setOreGround(int index, mp32_t amount);
    mp16_t getTemp() const;
    void   setTemp(mp16_t temp);

    const char* getTempStr() const
        { return getTempStr(getTemp()); }

    /* Accessors: Cargo */
    int32 getMoney() const
        { return full_info->data.money - cargo.getReserved(el_Money); }
    int32 getSupplies() const
        { return full_info->data.supplies - cargo.getReserved(el_Supplies); }
    int32 getCargo(GCargoType type) const;
    int32 getCargoRaw(GCargoType type) const;
    GCargo&  getCargoHandler()
        { return cargo; }
    void changeCargoRaw(GCargoType type, int32 amount);
    bool isCargoAmountKnown(GCargoType type) const;

    /* Accessors: Base specs */
    int getBaseDamage()  const
        { return base_dat->damage; }

    /* Accessors: Base mission */
    int   getBaseMission() const
        { return base_dat->mission; }
    const char* getBaseMissionName() const
        { return getBaseMissionName(getBaseMission()); }
    void  setBaseMission(int mission);

    /* Accessors: Shipyard */
    int   getBaseShipyardOrder() const
        { return base_dat->shipyard_action; }
    int   getBaseShipyardId() const
        { return base_dat->shipyard_id; }
    void  setBaseShipyardOrder(int action, int id);

    /* Accessors: Tech */
    int   getBaseTech(TTechLevel level) const
        { return base_dat->tech_levels[level]; }
    void  setBaseTech(TTechLevel level, int value);

    /* Accessors: build order. FIXME: THIS IS INCOMPLETE */
    int          getBaseBuildHull() const;
    TBuildOrder  getBaseBuildOrder() const;
    void         setBaseBuildOrder(const TBuildOrder& order);
    int          getShipCloningHere(const GUniverse& trn) const;

    int          getBaseEngineStore(int slot) const
        { return base_dat->engine_store[slot-1]; }
    int          getBaseBeamStore(int slot) const
        { return base_dat->beam_store[slot-1]; }
    int          getBaseHullStoreSlot(int slot) const
        { return base_dat->hull_store[slot-1]; }
    int          getBaseLauncherStore(int slot) const
        { return base_dat->launcher_store[slot-1]; }

    void         setBaseEngineStore(int slot, int amount);
    void         setBaseBeamStore(int slot, int amount);
    void         setBaseHullStoreSlot(int slot, int amount);
    void         setBaseLauncherStore(int slot, int amount);

    /* Administrivia */
    void internalCheck();
    void combinedCheck2(GUniverse& univ);

    void setShipInOrbit(const GShip& s);

    static bool isClippedCoord(GPoint p);

    enum {
        ps_DirtyHead = 1,
        ps_DirtyData = 2,
    };

    static const char* getTempStr(int temp);
    static const char* getNativeRaceName(int race);
    static const char* getNativeGovernmentName(int gov);
    static const char* getHappyStr(int happy);
    static const char* getChangeStr(int change);
    static const char* getBaseMissionName(int mission);
    static int32       getNativeDue(int tax, int race, int gov, int32 pop, int owner);
    static int32       getAmorphousBreakfast(int happy);
    static int32       getBovinoidSupplyContribution(int32 pop, int owner);

    enum Facility {
        /** Construction yard on starbase. This facility includes tech
            levels and starbase storage. */
        pf_BaseConstruction,
        /** Structures. This facility includes planetary structures
            and starbase defense. */
        pf_Structures,
        /** Base building. This facility describes the
            build-a-starbase order. */
        pf_BuildBase
    };
};

class GPlanetType : public GObjectType {
 public:
    GPlanetType(GUniverse& univ);

    bool isExistant(int index);
    GPlanet& getObject(int index);
    string_t getClassName();
    GPlanet& getPlanet()
        { return getObject(getCurrent()); }
};

class GBaseType : public GObjectType {
 public:
    GBaseType(GUniverse& univ);

    bool isExistant(int index);
    GPlanet& getObject(int index);
    string_t getClassName();
    GPlanet& getPlanet()
        { return getObject(getCurrent()); }
};

class GPlanetTransfer : public GCargoContainer {
    GPlanet& planet;
 protected:
    int32  getAmount(GCargoType type);
    void   changeAmount(GCargoType type, int32 amount);
    void   reserveAmount(GCargoType type, int32 amount);
 public:
    GPlanetTransfer(GPlanet& planet);
    ~GPlanetTransfer();

    bool     canHave(GCargoType cargo_id);
    int32    getMax(GCargoType cargo_id);
    string_t getName() const;

    GPlanet& getPlanet() { return planet; }
};

#endif
