/**
  *  \file gfx/sprite.h
  *  \brief Sprites
  *
  *  (c) 2001-2004 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_SPRITE_H
#define PCC_V2_GFX_SPRITE_H

#include <vector>
#include "gfx/gfxutil.h"
#include "gfx/gfx.h"
#include "gfx/pixmap.h"

class GfxSpriteManagerData;
class GfxSpriteManager;

//! Sprite.
class GfxSprite : public Object {
    friend class GfxSpriteManager;
    Ptr<GfxSpriteManagerData> mgr;
    Ptr<GfxSprite> next, prev;
    GfxSpriteManager* manager;
 protected:
    GfxRect pos;
    int layer;
 public:
    GfxSprite(const GfxRect& apos, int alayer);
    ~GfxSprite();
    void moveTo(int x, int y);
    void moveToC(int x, int y);
    void setPos(GfxRect new_pos);
    void update();
    void deleteThis();
    virtual void tick();

    /** Draw this sprite. This only needs to draw the sprite, update
        is automatic. */
    virtual void draw(GfxCanvas& can) = 0;

    void setManager(GfxSpriteManager* amanager);
    /** Get sprite manager associated with this sprite. */
    GfxSpriteManager* getManager() const
        { return manager; }

    /** Get position/size of this sprite. */
    const GfxRect& getPos() const
        { return pos; }

    /** Get center X position. */
    int getPosCX() const
        { return pos.x + pos.w / 2; }
    /** Get center Y position. */
    int getPosCY() const
        { return pos.y + pos.h / 2; }
    /** Get height of this sprite. */
    int getHeight() const
        { return pos.h; }
    /** Get width of this sprite. */
    int getWidth() const
        { return pos.w; }
    /** Get layer of this sprite. */
    int getLayer() const
        { return layer; }
    /** Get associated sprite manager data. */
    GfxSpriteManagerData* getManagerData() const
        { return mgr; }
};

//! Sprite Manager (container).
class GfxSpriteManager {
    std::vector<Ptr<GfxSprite> > sprite_list;
    GfxRect bbox;
 public:
    GfxSpriteManager();
    ~GfxSpriteManager();
    
    void addSprite(Ptr<GfxSprite> asprite);
    void tick(GfxCanvas& can);
    void tickSkip(bool call_tick = true);

    void moveSprite(Ptr<GfxSprite> asprite, int ax, int ay);
    void deleteSprite(Ptr<GfxSprite> asprite);
    void deleteLayer(int layer);

    void draw(GfxCanvas& can);

    void drawVoid(GfxCanvas& can, GfxRect r);

    /** Set bounding box. Output will be limited to this area. */
    void setBoundingBox(const GfxRect& r)
        { bbox = r; }
    /** Get current bounding box. \see setBoundingBox */
    const GfxRect& getBoundingBox() const
        { return bbox; }

    bool isLayerEmpty(int layer) const;
};

class GfxPixmapSprite : public GfxSprite {
    Ptr<GfxPixmap> pix;
 public:
    GfxPixmapSprite(int x, int y, int l, Ptr<GfxPixmap> apix);
    GfxPixmapSprite(int x, int y, int l, Ptr<GfxPixmap> apix, bool);
    void draw(GfxCanvas& can);
};

#endif
